@@ -24,7 +24,7 @@
   * @author Samuel Mimram
   *)
 
-(* $Id: ogg2wav.ml 3752 2007-05-23 20:40:34Z smimram $ *)
+(* $Id: ogg2wav.ml 5949 2008-10-02 13:00:45Z metamorph68 $ *)
 
 let bufsize = 16 * 1024
 
@@ -95,11 +95,11 @@ let _ =
       exit 1
     );
 
-  let df = Vorbis.Decoder.openfile !src in
-  let infos = Vorbis.Decoder.info df (-1) in
-  let vdr, cmt = Vorbis.Decoder.comments df (-1) in
-  let duration = Vorbis.Decoder.duration df (-1) in
-  let samples = Vorbis.Decoder.samples df (-1) in
+  let df = Vorbis.File.Decoder.openfile !src in
+  let infos = Vorbis.File.Decoder.info df (-1) in
+  let vdr, cmt = Vorbis.File.Decoder.comments df (-1) in
+  let duration = Vorbis.File.Decoder.duration df (-1) in
+  let samples = Vorbis.File.Decoder.samples df (-1) in
     Printf.printf
       "Input file characteristics: vorbis codec v%d, %d channels, %d Hz, %.02f s, %d samples\n"
       infos.Vorbis.vorbis_version
@@ -120,14 +120,14 @@ let _ =
           try
             while true
             do
-              let r = Vorbis.Decoder.decode df buf 0 bufsize in
+              let r = Vorbis.File.Decoder.decode df buf 0 bufsize in
                 output oc buf 0 r;
                 pos := !pos + r;
                 progress_bar "Decoding ogg:" (!pos / 4) tot
             done;
-            close_out oc; Vorbis.Decoder.close df
+            close_out oc; Vorbis.File.Decoder.close df
           with
-            | End_of_file -> close_out oc; Vorbis.Decoder.close df
+            | End_of_file -> close_out oc; Vorbis.File.Decoder.close df
       );
       Printf.printf "\n";
 
