@@ -49,6 +49,9 @@ static void raise_err(int ret)
 {
   switch(ret)
   {
+    case OV_FALSE:
+      caml_raise_constant(*caml_named_value("vorbis_exn_false"));
+
     case OV_HOLE:
       caml_raise_constant(*caml_named_value("vorbis_exn_hole_in_data"));
 
@@ -76,6 +79,9 @@ static void raise_err(int ret)
     case OV_EIMPL:
       caml_raise_constant(*caml_named_value("vorbis_exn_not_implemented"));
 
+    case OV_ENOTAUDIO:
+      caml_raise_constant(*caml_named_value("vorbis_exn_not_audio"));
+
     default:
       caml_raise_with_arg(*caml_named_value("vorbis_exn_unknown_error"), Val_int(ret));
   }
@@ -87,42 +93,221 @@ static void check_err(int ret)
     raise_err(ret);
 }
 
+/**** Decoding *****/
 
-/***** Decoding *****/
-
-
-/*
-CAMLprim value ocaml_vorbis_feed(value sync_state, value buffer, value _offs, value _len)
+typedef struct
 {
-  ogg_sync_state *oy = Sync_state_val(sync_state);
-  int offs = Int_val(_offs);
+  vorbis_dsp_state vd;
+  vorbis_block vb;
+  vorbis_info vi;
+  vorbis_comment vc;
+} decoder_t;
+
+#define Decoder_val(v) (*((decoder_t**)Data_custom_val(v)))
+#define Decoder_dsp_state_val(v) (&Decoder_val(v)->vd)
+#define Decoder_info_val(v) (&Decoder_val(v)->vi)
+#define Comment_val(v) (&Decoder_val(v)->vc)
+#define Block_val(v) (&Decoder_val(v)->vb)
+
+static void finalize_decoder(value e)
+{
+  decoder_t *dec = Decoder_val(e);
+
+  vorbis_block_clear(&dec->vb);
+  vorbis_dsp_clear(&dec->vd);
+  vorbis_info_clear(&dec->vi);
+  vorbis_comment_clear(&dec->vc);
+  free(dec);
+}
+
+static struct custom_operations decoder_ops =
+{
+  "ocaml_decoder_t",
+  finalize_decoder,
+  custom_compare_default,
+  custom_hash_default,
+  custom_serialize_default,
+  custom_deserialize_default
+};
+
+CAMLprim value ocaml_vorbis_val_info_of_decoder(value vorbis_t)
+{
+  CAMLparam1(vorbis_t);
+  CAMLlocal1 (v);
+  int i = 0;
+  v = caml_alloc_tuple(7);
+  vorbis_info *vi = Decoder_info_val(vorbis_t);
+
+  Store_field (v, i++, Val_int(vi->version));
+  Store_field (v, i++, Val_int(vi->channels));
+  Store_field (v, i++, Val_long(vi->rate));
+  Store_field (v, i++, Val_long(vi->bitrate_upper));
+  Store_field (v, i++, Val_long(vi->bitrate_nominal));
+  Store_field (v, i++, Val_long(vi->bitrate_lower));
+  Store_field (v, i++, Val_long(vi->bitrate_window));
+
+  CAMLreturn(v);
+}
+
+CAMLprim value ocaml_vorbis_val_comments_of_decoder(value decoder)
+{
+  CAMLparam1(decoder);
+  CAMLlocal2(ans, cmts);
+
+  int i;
+  vorbis_comment *vc = Comment_val(decoder);
+
+  cmts = caml_alloc_tuple(vc->comments);
+  for (i = 0; i < vc->comments; i++)
+    Store_field(cmts, i, caml_copy_string(vc->user_comments[i]));
+  ans = caml_alloc_tuple(2);
+  if (vc->vendor != NULL)
+    Store_field(ans, 0, caml_copy_string(vc->vendor));
+  else
+    Store_field(ans, 0, caml_copy_string("(null)"));
+  Store_field(ans, 1, cmts);
+
+  CAMLreturn(ans);
+}
+
+CAMLprim value ocaml_vorbis_check_packet(value packet)
+{
+  CAMLparam1(packet);
+  ogg_packet *op = Packet_val(packet);
+  vorbis_info vi;
+  vorbis_comment vc;
+  vorbis_info_init(&vi);
+  vorbis_comment_init(&vc);
+  int ret = 1;
+  if(vorbis_synthesis_headerin(&vi, &vc, op) < 0)
+    ret = 0;
+  vorbis_info_clear(&vi);
+  vorbis_comment_clear(&vc);
+  CAMLreturn(Val_int(ret));
+}
+
+CAMLprim value ocaml_vorbis_synthesis_init(value packet, value packet2, value packet3)
+{
+  CAMLparam3(packet,packet2,packet3);
+  CAMLlocal1(ans);
+  ogg_packet *op = Packet_val(packet);
+  ogg_packet *op2 = Packet_val(packet2);
+  ogg_packet *op3 = Packet_val(packet3);
+  int ret;
+
+  decoder_t *vt = malloc(sizeof(decoder_t));
+  if (vt == NULL)
+    caml_failwith("malloc");
+
+  vorbis_info_init(&vt->vi);
+  vorbis_comment_init(&vt->vc);
+
+  ret = vorbis_synthesis_headerin(&vt->vi, &vt->vc, op);
+  if(ret < 0)
+  {
+    vorbis_info_clear(&vt->vi);
+    vorbis_comment_clear(&vt->vc);
+    free(vt);
+    raise_err(ret);
+  }
+  ret = vorbis_synthesis_headerin(&vt->vi, &vt->vc, op2);
+  if(ret < 0)
+  {
+    vorbis_info_clear(&vt->vi);
+    vorbis_comment_clear(&vt->vc);
+    free(vt);
+    raise_err(ret);
+  }
+  ret = vorbis_synthesis_headerin(&vt->vi, &vt->vc, op3);
+  if (ret < 0)
+  {
+    vorbis_info_clear(&vt->vi);
+    vorbis_comment_clear(&vt->vc);
+    free(vt);
+    raise_err(ret);
+  }
+
+  vorbis_synthesis_init(&vt->vd, &vt->vi);
+  vorbis_block_init(&vt->vd,&vt->vb);
+  
+  ans = caml_alloc_custom(&decoder_ops, sizeof(decoder_t*), 1, 0);
+  Decoder_val(ans) = vt;
+
+  CAMLreturn(ans);
+}
+
+CAMLprim value ocaml_vorbis_decode_pcm(value vorbis_state, value stream_state, value buf, value _pos, value _len)
+{
+  CAMLparam3(vorbis_state,stream_state,buf);
+  CAMLlocal2(buffer,chan);
+  ogg_stream_state *os = Stream_state_val(stream_state);
+  ogg_packet op;
+  vorbis_block *vb = Block_val(vorbis_state);
+  vorbis_dsp_state *vd = Decoder_dsp_state_val(vorbis_state);
+  vorbis_info *vi = Decoder_info_val(vorbis_state);
+  int pos = Int_val(_pos);
   int len = Int_val(_len);
-  char *outbuf = ogg_sync_buffer(oy, len);
+  float **pcm;
+  int samples;
+  int i,j,ret;
+  int total_samples = 0;
+ 
+  while (1) { 
+    while (total_samples < len) {
+      samples = vorbis_synthesis_pcmout(vd,&pcm);
+      if (samples < 0)
+        raise_err(samples);
+      if (samples == 0)
+        break;
+      if (samples > len - total_samples)
+        samples = len - total_samples;
+      if (Wosize_val(buf) < vi->channels)
+        caml_raise_constant(*caml_named_value("vorbis_exn_invalid"));
+      for (i=0 ; i<vi->channels ; i++)
+      {
+        chan = Field(buf,i);
+        if (Wosize_val(chan) / Double_wosize - pos < samples)
+          caml_raise_constant(*caml_named_value("vorbis_exn_invalid"));
+        for (j=0; j<samples; j++)
+          Store_double_field(chan,pos+j,pcm[i][j]);
+      }
+      pos += samples;
+      total_samples += samples;
+      ret = vorbis_synthesis_read(vd,samples);
+      if (ret < 0)
+        raise_err(ret);
+    }
+    if (total_samples == len)
+      CAMLreturn(Val_int(total_samples));
 
-  memcpy(outbuf, String_val(buffer) + offs, len);
-  ogg_sync_wrote(oy, len);
+    ret = ogg_stream_packetout(os,&op);
+    if (ret == 0)
+      CAMLreturn(Val_int(total_samples)); 
+    if (ret < 0)
+      raise_err(ret);
 
-  return Val_unit;
+    caml_enter_blocking_section();
+    ret = vorbis_synthesis(vb,&op);
+    caml_leave_blocking_section();
+
+    if (ret == 0)
+      ret = vorbis_synthesis_blockin(vd,vb);
+    if (ret < 0)
+      raise_err(ret);
+  }
+
+  CAMLreturn(Val_int(total_samples));
 }
 
-CAMLprim value ocaml_vorbis_synthesis_init(value unit)
+CAMLprim value ocaml_vorbis_synthesis_restart(value s)
 {
-  vorbis_info *vi = malloc(sizeof(vorbis_info));
-  value ans;
-
-  vorbis_info_init(vi);
-  // TODO!
-  ans = new_dsp_state();
-  vorbis_analysis_init(Dsp_state_val(ans), vi);
-
-  return ans;
+  CAMLparam1(s);
+  vorbis_synthesis_restart(Decoder_dsp_state_val(s));
+  CAMLreturn(Val_unit);
 }
-*/
-
 
 /***** Encoding *****/
 
-
 typedef struct
 {
   vorbis_dsp_state vd;
@@ -132,7 +317,7 @@ typedef struct
 } encoder_t;
 
 #define Encoder_val(v) (*((encoder_t**)Data_custom_val(v)))
-#define Dsp_state_val(v) (&Encoder_val(v)->vd)
+#define Enc_dsp_state_val(v) (&Encoder_val(v)->vd)
 
 static void finalize_encoder(value e)
 {
@@ -200,10 +385,11 @@ CAMLprim value ocaml_vorbis_reset(value vdsp)
   return Val_unit;
 }
 
-CAMLprim value ocaml_vorbis_analysis_headerout(value vdsp, value vogg, value comments)
+CAMLprim value ocaml_vorbis_analysis_headerout(value vdsp, value comments)
 {
-  vorbis_dsp_state *vd = Dsp_state_val(vdsp);
-  ogg_stream_state *os = Stream_state_val(vogg);
+  CAMLparam2(vdsp, comments);
+  CAMLlocal4(ret,p1,p2,p3);
+  vorbis_dsp_state *vd = Enc_dsp_state_val(vdsp);
   vorbis_comment vc;
   ogg_packet header, header_comm, header_code;
   int i;
@@ -213,11 +399,13 @@ CAMLprim value ocaml_vorbis_analysis_headerout(value vdsp, value vogg, value com
     vorbis_comment_add_tag(&vc, String_val(Field(Field(comments, i), 0)), String_val(Field(Field(comments, i), 1)));
   vorbis_analysis_headerout(vd, &vc, &header, &header_comm, &header_code);
   vorbis_comment_clear(&vc);
-  ogg_stream_packetin(os, &header);
-  ogg_stream_packetin(os, &header_comm);
-  ogg_stream_packetin(os, &header_code);
 
-  return Val_unit;
+  ret = caml_alloc_tuple(3);
+  Store_field(ret,0,value_of_packet(&header));
+  Store_field(ret,1,value_of_packet(&header_comm));
+  Store_field(ret,2,value_of_packet(&header_code));
+
+  CAMLreturn(ret);
 }
 
 CAMLprim value ocaml_vorbis_encode_float(value vdsp, value vogg, value data, value _offs, value _len)
@@ -225,7 +413,7 @@ CAMLprim value ocaml_vorbis_encode_float(value vdsp, value vogg, value data, val
   CAMLparam3(vdsp, vogg, data);
   encoder_t *enc = Encoder_val(vdsp);
   vorbis_block *vb = &enc->vb;
-  vorbis_dsp_state *vd = Dsp_state_val(vdsp);
+  vorbis_dsp_state *vd = Enc_dsp_state_val(vdsp);
   ogg_stream_state *os = Stream_state_val(vogg);
   ogg_packet *op = &enc->op;
   int offs = Int_val(_offs);
@@ -264,8 +452,15 @@ CAMLprim value ocaml_vorbis_encode_float(value vdsp, value vogg, value data, val
   CAMLreturn(Val_unit);
 }
 
+CAMLprim value ocaml_vorbis_encode_time_of_granulepos(value v_state, value gpos)
+{
+  CAMLparam2(v_state,gpos);
+  encoder_t *enc = Encoder_val(v_state);
+  ogg_int64_t granulepos = Int64_val(gpos);
+  CAMLreturn(caml_copy_nativeint(vorbis_granule_time(&enc->vd,granulepos)));
+}
 
-/***** Decoding *****/
+/***** File Decoding *****/
 
 
 /* This should be malloced since we might want to register *_func as global root. */
@@ -285,7 +480,22 @@ static void finalize_dec_file(value _df)
 {
   myvorbis_dec_file_t *df = Decfile_val(_df);
 
-  // TODO: check that close has been done, otherwise df->ovh leaks
+  if (df->ovf)
+  {
+    caml_enter_blocking_section();
+    ov_clear(df->ovf);
+    caml_leave_blocking_section();
+    free(df->ovf);
+    df->ovf=NULL;
+  }
+  if(df->read_func)
+  {
+    caml_remove_global_root(&df->read_func);
+    caml_remove_global_root(&df->seek_func);
+    caml_remove_global_root(&df->close_func);
+    caml_remove_global_root(&df->tell_func);
+    df->read_func=0;
+  }
 
   free(df);
 }
@@ -620,7 +830,10 @@ CAMLprim value ocaml_vorbis_get_dec_file_comments(value d_f, value link_)
   for (i = 0; i < vc->comments; i++)
     Store_field(cmts, i, caml_copy_string(vc->user_comments[i]));
   ans = caml_alloc_tuple(2);
-  Store_field(ans, 0, caml_copy_string(vc->vendor));
+  if (vc->vendor != NULL)
+    Store_field(ans, 0, caml_copy_string(vc->vendor));
+  else
+    Store_field(ans, 0, caml_copy_string("(null)"));
   Store_field(ans, 1, cmts);
 
   CAMLreturn(ans);
@@ -694,3 +907,69 @@ CAMLprim value ocaml_vorbis_decoder_serialnumber(value d_f, value bs)
 
   CAMLreturn(Val_int(ret));
 }
+
+/* Ogg skeleton interface */
+
+/* Wrappers */
+static void write32le(unsigned char *ptr,ogg_uint32_t v)
+{
+  ptr[0]=v&0xff;
+  ptr[1]=(v>>8)&0xff;
+  ptr[2]=(v>>16)&0xff;
+  ptr[3]=(v>>24)&0xff;
+}
+
+static void write64le(unsigned char *ptr,ogg_int64_t v)
+{
+  ogg_uint32_t hi=v>>32;
+  ptr[0]=v&0xff;
+  ptr[1]=(v>>8)&0xff;
+  ptr[2]=(v>>16)&0xff;
+  ptr[3]=(v>>24)&0xff;
+  ptr[4]=hi&0xff;
+  ptr[5]=(hi>>8)&0xff;
+  ptr[6]=(hi>>16)&0xff;
+  ptr[7]=(hi>>24)&0xff;
+}
+
+/* Values from http://xiph.org/ogg/doc/skeleton.html */
+#define FISBONE_IDENTIFIER "fisbone\0"
+#define FISBONE_MESSAGE_HEADER_OFFSET 44
+#define FISBONE_SIZE 52
+
+/* Code from theorautils.c in ffmpeg2theora */
+CAMLprim value ocaml_vorbis_skeleton_fisbone(value serial, value samplerate, value start, value content)
+{
+  CAMLparam4(serial,samplerate,start,content);
+  CAMLlocal1(packet);
+  ogg_packet op;
+  int len = FISBONE_SIZE+caml_string_length(content);
+
+  memset (&op, 0, sizeof (op));
+  op.packet = malloc(len);
+  if (op.packet == NULL)
+    caml_failwith("malloc");
+
+  memset (op.packet, 0, len);
+  /* it will be the fisbone packet for the vorbis audio */
+  memcpy (op.packet, FISBONE_IDENTIFIER, 8); /* identifier */
+  write32le(op.packet+8, FISBONE_MESSAGE_HEADER_OFFSET); /* offset of the message header fields */
+  write32le(op.packet+12, Nativeint_val(serial)); /* serialno of the vorbis stream */
+  write32le(op.packet+16, 3); /* number of header packet */
+  /* granulerate, temporal resolution of the bitstream in Hz */
+  write64le(op.packet+20, (ogg_int64_t)Int64_val(samplerate)); /* granulerate numerator */
+  write64le(op.packet+28, (ogg_int64_t)1); /* granulerate denominator */
+  write64le(op.packet+36, (ogg_int64_t)Int64_val(start)); /* start granule */
+  write32le(op.packet+44, 2); /* preroll, for vorbis its 2 */
+  *(op.packet+48) = 0; /* granule shift, always 0 for vorbis */
+  memcpy (op.packet+FISBONE_SIZE, String_val(content), caml_string_length(content));
+
+  op.b_o_s = 0;
+  op.e_o_s = 0;
+  op.bytes = len;
+
+  packet = value_of_packet(&op);
+  free(op.packet);
+  CAMLreturn(packet);
+}
+
